/*******************************************************************************
Copyright (C) Marvell International Ltd. and its affiliates

********************************************************************************
Marvell GPL License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File in accordance with the terms and conditions of the General 
Public License Version 2, June 1991 (the "GPL License"), a copy of which is 
available along with the File in the license.txt file or by writing to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 or 
on the worldwide web at http://www.gnu.org/licenses/gpl.txt. 

THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE IMPLIED 
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY 
DISCLAIMED.  The GPL License provides additional details about this warranty 
disclaimer.

*******************************************************************************/

/*
 * serial.c - serial support.
 */

#include <common.h>
#include <command.h>
#include "mvUart.h"


extern void print_mvBanner(void);
extern void print_dev_id(void);


int serial_init (void)
{
	DECLARE_GLOBAL_DATA_PTR;
	
	int clock_divisor = (CFG_TCLK / 16)/gd->baudrate;
#if defined(CONFIG_KUROBOX_PLATFORM) && !defined(ORG_PCB)
	int clock_divisor2 = (CFG_TCLK / 16)/ 38400 ;
#endif

#ifdef CFG_INIT_CHAN1
	mvUartInit(0, clock_divisor, mvUartBase(0));
#endif
#ifdef CFG_INIT_CHAN2
 #if defined(CONFIG_KUROBOX_PLATFORM) && !defined(ORG_PCB)
 	mvUartInit(1, clock_divisor2, mvUartBase(1));
 #else
 	mvUartInit(1, clock_divisor, mvUartBase(1));
 #endif
#endif
        /* print banner */
        print_mvBanner();
        print_dev_id();
#if defined(CONFIG_KUROBOX_PLATFORM)
/*		fl_ok();*/
#endif

	return (0);
}

void
serial_putc(const char c)
{
	if (c == '\n')
		mvUartPutc(CFG_DUART_CHAN, '\r');

	mvUartPutc(CFG_DUART_CHAN, c);
}

int
serial_getc(void)
{	
	
	return mvUartGetc(CFG_DUART_CHAN);
}

int
serial_tstc(void)
{
	return mvUartTstc(CFG_DUART_CHAN);
}

void
serial_setbrg (void)
{
	DECLARE_GLOBAL_DATA_PTR;
	
	int clock_divisor = (CFG_TCLK / 16)/gd->baudrate;
#if defined(CONFIG_KUROBOX_PLATFORM) && !defined(ORG_PCB)
	int clock_divisor2 = (CFG_TCLK / 16)/ 38400 ;
#endif

#ifdef CFG_INIT_CHAN1
	mvUartInit(0, clock_divisor, mvUartBase(0));
#endif
#ifdef CFG_INIT_CHAN2
 #if defined(CONFIG_KUROBOX_PLATFORM) && !defined(ORG_PCB)
	mvUartInit(1, clock_divisor2, mvUartBase(1));
 #else
	mvUartInit(1, clock_divisor, mvUartBase(1));
 #endif
#endif
}


void
serial_puts (const char *s)
{
	while (*s) {
		serial_putc (*s++);
	}
}

#if (CONFIG_COMMANDS & CFG_CMD_KGDB)
void
kgdb_serial_init(void)
{
}

void
putDebugChar (int c)
{
	serial_putc (c);
}

void
putDebugStr (const char *str)
{
	serial_puts (str);
}

int
getDebugChar (void)
{
	return serial_getc();
}

void
kgdb_interruptible (int yes)
{
	return;
}
#endif	/* CFG_CMD_KGDB	*/
